#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1994/01/29 07:47:11 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1994/01/29 07:47:11 $
 *
 * The rtfifo object, which maintains a bidirectional fifo link
 * between two CPs (real or simulated.)  In simulation it's done
 * with a pair of pipes.
 * 
 * Each CP maintains a rtfifo for each other CP.  rtfifos are bidirectional.
 * 
 * sched advance?
 * vector size for sigs?
 * 
 * To read the input messages for a tick, call rtfifo_tickin().
 * To flush output messages and set up signal I/O call rtfifo_setupsigs().
 * To flush the output signals call rtfifo_donesigs().
 * The three calls must be made in that order, once per tick.
 * 
 * To send a message along an rtfifo, use rtfifo_sendmess();
 * to send a packet for the host (which should already be formatted
 * as printout or an update), use rtfifo_sendtohost() -- which had
 * better go to the master CP (number 0.)  Make these calls after calling
 * rtfifo_tickout() for the previous tick and before calling rtfifo_setupsigs()
 * for the current one.
 * 
 * service routines rtfifo_addsig() and rtfifo_rmsig() add or remove a signal
 * from the transmission list.  The receive list of the corresponding rtfifo
 * is updated accordingly.
 * 
 * The scheduling advance of the sending processor is taken to be SCHEDADVANCE
 * ticks -- beware, this is not necessarily what the a/d/a think.  Later,
 * routines should be added to change this number.
 * 
 * On receiving a send-to-host packet (which happens during rtfifo_getmess):
 * if the packet is for printout, it is sent to "post()", and if it's
 * a host queue update, it's kept and the rtfifo's own host queue element is
 * set.  As soon as the buffer contents are empty again a token is sent to
 * allow the other CP to send some other host queue update.
 * 
 * The message packet format (as physically written between CPs is:
 * 
 * [type] [id] [atom] ...
 * 
 * where type is one of CP_MESS, etc (defined in cp_mess.h), destination-id
 * is the host ID of the destination object
 * where an atom is:
 * 
 * char type;
 * union
 * {
 * long longval;
 * float floatval;
 * }
 * 
 * for a number and
 * 
 * char type;
 * char null-terminated-name[];
 * 
 * for a symbol.  Atoms are written to the pipe in (long; float) alignment,
 * that is, 4 byte alignment on the 860 (and most other processors today)).
 * 
 * As a special case, the destination-ID zero means, "anything else besides
 * a message in the above format"; the packet is:
 * 
 * 0 [type] [anything].
 * 
 * The type can be:
 * 
 * H_ENDOFTICK          (to mark the last message packet of a tick)
 * H_PRINT [size] [data]    (send printout to host)
 * H_QUEUE [size] [data]    (send update to host)
 * H_SETNVEC [nvecs]      (to change the number of signal vectors sent)
 * 
 * 
 * Note that H_PRINT and H_QUEUE are in exactly the same format that the host
 * requires, except that the header and size must be put in the host's byte order.
 * 
 * For real-time, an exotic circular buffer pool packing format is planned but
 * not implemented.  In simulation, a rtfifo maintains a pipe pair.
 * 
 * To avoid pipe deadlock, the writing end of the pipes are written
 * fail-if-would-block.  On failure, we select on any input pipe being ready
 * until the write can go through, squirreling the read results away for later.
 * (can you select on an output pipe being writable???)
 * 
 * Problem: you can choke everything by setting up a circular message path
 * with "gain".  There could be some sanity check on the number of messages
 * that may be sent per tick, or per SCHEDADVANCE-tick period, perhaps.
 * 
 * Still doesn't shut down gracefully in simulation.
 *
 */

#include "m_mess.h"
#include "cp_mess.h"

#ifdef I860
#include "../6a/6a.h"
#include "../6a/proto56.h"
void *sys_slottoaddr();

#define FIFOADDR  (IOSPACE & 0x7f0000)  /* input buffers -- grow down */
extern int cp_slotno;
#else
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/select.h>
#endif                          /* I860 */
void rtfifo_setupsigs(), rtfifo_sigsin(), rtfifo_sigsout();


extern void *h_objlist;
extern int cp_this;             /* ID number of this CP */
extern int cp_nproc;            /* number of CPs */
extern void *u_chain;           /* used as a flag saying whether DSP is on */

static int rtfifo_blowoff;      /* flag to break out of input or output polling */
static int rtfifo_sigsused;     /* flag to inhibit flushing output buffer */
static int rtfifo_cpfrom;       /* cp that read current message */
static int rtfifo_outtickcount = 1;
static int rtfifo_advance;
static int rtfifo_deltaadvance;

/* Constant for the rtfifo */

#define IPCVS 64                /* vector size for a/d/a and inter-CP signal paths */
#define OUTBUFSIZE 4096         /* output buffer size (a multipe of IPC signal size) */

#ifdef I860
/* Fuck the compiler !*/
static int _inBufSize = 65536;
#define INBUFSIZE _inBufSize    /* input buffer size (a multiple of OUTBUFSIZE) */
#else
#define INBUFSIZE 262144        /* bigger in simulation */
#endif

#define SIGALIGN 16             /* signal alignment */
#define SIGSIZE (IPCVS * sizeof(float)) /* size of a signal */
#define MAXTRANSSYM 4096        /* truncate huge symbols */
#define MAXMESSARGCOUNT 256     /* truncate long messages */
/* max number of inter-CP sigs: */
#define MAXOUTSIG (OUTBUFSIZE/(IPCVS * sizeof(float)) - 1)

typedef struct sigio {
    int s_cpno;
    int s_write;
    float **s_where;
    struct sigio *s_next;
} t_sigio;

typedef struct rtfifo {
    int f_cpno;                 /* ID of CP we're connected to */
#ifdef I860
    t_ftscpu *f_lctladdr;       /* our ftscpu entry for other CP */
    char *f_rbufaddr;           /* other CP's input buffer address */
    t_ftscpu *f_rctladdr;       /* other CP's "ftscpu" entry for this one */
    int f_remotehead;           /* copy of remote head */
    int f_dmathresh;            /* minimum number of bytes to make DMA worth it */
#endif
    /* input */
#ifndef I860
    int f_infd;                 /* filedesc for input */
#endif
    char *f_inbuf;              /* fifo buffer */
    int f_inhead;               /* fifo head (onset of first empty byte) */
    int f_intail;               /* fifo tail (onset of first full byte) */
    int f_nexttick;             /* next tick number expected from other CP */
    /* eat ticks from other CP (or if negative, */
    /* number of extra ticks to eat right away) */
    int f_inadvance;            /* scheduler advance of other CP */
    int f_insigcount;           /* number of signals coming from other CP */
    /* output */
#ifndef I860
    int f_outfd;                /* output file descriptor */
    int f_logfd;                /* logfile fd */
#endif
    char *f_outmessbuf;         /* message output buffer */
    char *f_outmesshead;        /* first empty byte in output buffer */
    char *f_outmessend;         /* end of output buffer */
    int f_byteswritten;         /* total bytes ever written out the door */
    int f_outsigcount;          /* number of signals going to other CP */
} t_rtfifo;

t_rtfifo *rtfifo_vec[MAXCP];    /* pointers to all rtfifos */
void *malloc();

/* debugging messages go somewhere safe, not over the protocol! */
#ifdef I860
rtfifo_debug(s, a, b, c, d, e)
char *s;
#else
rtfifo_debug(char *s, int a, int b, int c, int d, int e)
#endif
{
#ifdef I860
    sa_printf("\n%d: ", cp_this);
    sa_printf(s, a, c, b, d, e);
#else
    fprintf(stderr, "\n%d: ", cp_this);
    fprintf(stderr, s, a, c, b, d, e);
#endif                          /* I860 */
}

rtfifo_die(s, a, b, c, d, e)
char *s;
int a;
int b;
int c;
int d;
int e;
{
    rtfifo_debug(s, a, b, c, d, e);
#ifdef I860
    sa_printf("rtfifo: fatal error\n");
    sys_spin(2);
#else
    hate("\nfailure in inter-CP mechanism");
#endif                          /* I860 */
}

t_sigio *sig_inlist, *sig_outlist, *sig_inptr, *sig_outptr;

void sigio_free(t_sigio *x)
{
    t_sigio stub;
    t_sigio *sp, *sp2;
    stub.s_next = (x->s_write ? sig_outlist : sig_inlist);
    for (sp = &stub; sp2 = sp->s_next; sp = sp2)
    {
        if (sp2 == x)
        {
            sp->s_next = sp2->s_next;
            goto didit;
        }
    }
    rtfifo_die("sigout_free");
  didit:
    if (x->s_write)
    {
        sig_outlist = stub.s_next;
        rtfifo_vec[x->s_cpno]->f_outsigcount--;
    }
    else
    {
        sig_inlist = stub.s_next;
        rtfifo_vec[x->s_cpno]->f_insigcount--;
    }
    freebytes(x, sizeof(t_sigio));
}

void sigio_relocate(t_sigio *x, float **where)
{
    x->s_where = where;
}

/* create a new one.  Cpno is only used for a "write" sigio; a read one is
   always set up as a result of a remote message being received and we
   recover the sending CP number. */

void *sigio_new(float **where, int wflag, int cpno)
{
    t_sigio *x = (t_sigio *) getbytes(sizeof(t_sigio));
    x->s_where = where;
    x->s_write = wflag;
    if (x->s_write)
    {
        if (cpno < 0 || cpno >= cp_nproc || cpno == cp_this)
            rtfifo_die("rtfifo_new");
        x->s_cpno = cpno;
        x->s_next = sig_outlist;
        sig_outlist = x;
        /* this should resize the output buffer to be able to 
           hold the new signals; instead we just crash and burn. */
        if (++(rtfifo_vec[x->s_cpno]->f_outsigcount) > MAXOUTSIG)
            hate("too many signals passed between CPs %d and %d (%d maximum)", cp_this + 1, x->s_cpno + 1, MAXOUTSIG);
    }
    else
    {
        x->s_cpno = rtfifo_cpfrom;
        x->s_next = sig_inlist;
        sig_inlist = x;
        /* this can be deleted when resizing is fixed */
        if (++(rtfifo_vec[x->s_cpno]->f_insigcount) > MAXOUTSIG)
            hate("too many signals passed between CPs %d and %d (%d maximum)", x->s_cpno + 1, cp_this + 1, MAXOUTSIG);
    }
    return (x);
}

/*  It would be nice to be able to change the inter-CP signal vector size
    dynamically ...  At this point you can't even change the number of
    samples in a DSP tick since each tick has to send out a signal buffer
    set.  You could at least hope to get submultiples of DSP tick size. */

void rtfifo_setvs(int n)
{
    if (n != IPCVS)
        rtfifo_die("rtfifo_setvs: can't change vector size");
}

/*  Stop the bus.  Closes all outputs and reads all inputs until EOF.  You
  may call this twice if you want; e.g. once from this file when an I/O
  problem comes up and again from main as part of the cleanup. */

void rtfifo_shutdown()
{
#ifndef I860
    t_rtfifo *x;
    int i;
    if (rtfifo_blowoff)
        return;
    post("shut down");
    for (i = 0; i < cp_nproc; i++)
        if (i != cp_this)
            post("close %d", rtfifo_vec[i]->f_outfd),
                close(rtfifo_vec[i]->f_outfd);
    while (1)
    {
        fd_set readset, writeset, exceptset;
        int i, maxfd = 0, sel;
        FD_ZERO(&readset);
        FD_ZERO(&writeset);
        FD_ZERO(&exceptset);
        for (i = 0; i < cp_nproc; i++)
            if (i != cp_this)
            {
                x = rtfifo_vec[i];
                if (x->f_infd >= 0)
                {
                    int fd = x->f_infd;
                    FD_SET(fd, &readset);
                    if (fd >= maxfd)
                        maxfd = fd + 1;
                }
            }
        if (!maxfd)
            break;
        sel = select(maxfd + 1, &readset, &writeset, &exceptset, 0);
        if (sel < 0)
        {
            perror("rtfifo_shutdown");
            hate("select failed");
        }
        /* read any inputs that are ready */
        for (i = 0; i < cp_nproc; i++)
            if (i != cp_this
                && FD_ISSET((x = rtfifo_vec[i])->f_infd, &readset))
            {
                char buf[OUTBUFSIZE];
                int rd = read(x->f_infd, buf, OUTBUFSIZE);
                if (rd < 0)
                {
                    perror("rtfifo_shutdown");
                    hate("read failed");
                }
                else if (rd == 0)
                    x->f_infd = -1;
            }
    }
    rtfifo_blowoff = 1;
#endif
}

/*  Read more bytes into the input buffer, blockingly.  Returns 1
    and shuts down on failure. */

#define ITER_LIMIT 2

rtfifo_getmore(t_rtfifo *x)
{
#ifdef I860
    int inval;
    while ((inval = x->f_lctladdr->cpu_head) == x->f_inhead)
    {
        cp_spin(156);
        if (!PROTO->p_ftsflags)
        {
            int i;
            setLEDS(LED3, 16);
            for (i = 0; i < 10000000; i++);
            sys_spin(3);
        }
    }
    x->f_inhead = inval;
    return (0);
#else
    int r = (x->f_intail > x->f_inhead ? x->f_intail : INBUFSIZE) -
        x->f_inhead;
    int got;
    if (rtfifo_blowoff)
        return (1);
    if (r <= 0)
        rtfifo_die("rtfifo_getmore");
    got = read(x->f_infd, x->f_inbuf + x->f_inhead, r);
    if (got <= 0)
    {
        if (got < 0)
            perror("rtfifo_getmore");
        rtfifo_shutdown();
        return (1);
    }
    x->f_inhead += got;
    if (x->f_inhead == INBUFSIZE)
        x->f_inhead = 0;
    return (0);
#endif
}

/*  Flush the output buffer.  If you're using pipes, you have to do "select"
  on all the input buffers while you wait for the output to be sent.  (If
  there's input, it's stuffed in a buffer to be dealt with later).  If
  you're using shared memory you can just write straight to the other CP's
  buffer */


void rtfifo_flush(t_rtfifo *x)
{
#ifdef I860
    int bcount = x->f_outmesshead - x->f_outmessbuf;
    int usedma = (bcount > x->f_dmathresh);
    if (!bcount)
        return;
    if (bcount & (SIGALIGN - 1))
        rtfifo_die("flush alignment");

    if (x->f_remotehead & (SIGALIGN - 1))
        rtfifo_die("flush alignment 2");
    if (bcount + x->f_remotehead >= INBUFSIZE)  /* wrap? */
    {
        int size1 = INBUFSIZE - x->f_remotehead;
        if (size1 < 0 || size1 > INBUFSIZE)
            rtfifo_die("wraparound");
        io_cpy860(x->f_rbufaddr + x->f_remotehead,
                  x->f_outmessbuf, size1, usedma);
        if (bcount > size1)
            io_cpy860(x->f_rbufaddr, x->f_outmessbuf + size1,
                      bcount - size1, usedma);
        x->f_remotehead -= INBUFSIZE;
    }
    else
    {

        io_cpy860(x->f_rbufaddr + x->f_remotehead, x->f_outmessbuf,
                  bcount, usedma);

#ifdef UNDEFINED
        if (x->f_cpno == 4 &&
            *(long *) (x->f_rbufaddr + x->f_remotehead)
            != *(long *) (x->f_outmessbuf))
        {
            LEDS[LED2] = 16;
            sa_printf("copy failed: %lx -> %lx\n",
                      *(long *) (x->f_outmessbuf),
                      *(long *) (x->f_rbufaddr + x->f_remotehead));
            sa_printf("verify: %lx -> %lx\n",
                      *(long *) (x->f_outmessbuf),
                      *(long *) (x->f_rbufaddr + x->f_remotehead));
            sys_spin(4);
        }
#endif
    }
    x->f_remotehead += bcount;
    /* x->f_rctladdr->cpu_head = x->f_remotehead; (now done later) */
    x->f_byteswritten += bcount;
    x->f_outmesshead = x->f_outmessbuf;
    /* if (x->f_remotehead >= 16384) { rtfifo_debug("shit\n"); while (1); } */
    /* rtfifo_debug("flush to %lx\n", x->f_byteswritten); */
#else
    char *ebuf = x->f_outmesshead;
    char *where = x->f_outmessbuf;
    t_rtfifo *y;
    if ((x->f_outmesshead - x->f_outmessbuf) & (SIGALIGN - 1))
        rtfifo_die("flush alignment");
    /* post("flush %d", x->f_outmesshead - x->f_outmessbuf); */
    if (where == ebuf)
        return;
    if (rtfifo_sigsused)
    {
        rtfifo_die("choked during DSP loop");
        x->f_outmesshead = x->f_outmessbuf;
        return;
    }
    while (1)
    {
        fd_set readset, writeset, exceptset;
        int i, maxfd = 0, sel;
        if (rtfifo_blowoff)
        {
            x->f_outmesshead = where;
            return;
        }
        FD_ZERO(&readset);
        FD_ZERO(&writeset);
        FD_ZERO(&exceptset);
        for (i = 0; i < cp_nproc; i++)
            if (i != cp_this)
            {
                int fd = rtfifo_vec[i]->f_infd;
                FD_SET(fd, &readset);
                if (fd >= maxfd)
                    maxfd = fd + 1;
            }
        FD_SET(x->f_outfd, &writeset);
        if (x->f_outfd >= maxfd)
            maxfd = x->f_outfd + 1;
        sel = select(maxfd + 1, &readset, &writeset, &exceptset, 0);
        if (sel < 0)
        {
            perror("rtfifo_flush");
            hate("select failed");
        }
        /* try to write output */
        if (FD_ISSET(x->f_outfd, &writeset))
        {
            int w = write(x->f_outfd, where, ebuf - where);
            if (w <= 0)
            {
                if (w < 0)
                {
                    perror("rtfifo_flush");
                    hate("select failed");
                }
                else
                    rtfifo_die("write returned zero???");
            }
            if (x->f_logfd >= 0)
                write(x->f_logfd, where, w);
            /* post("write %d %d", where - x->f_outmessbuf,  w); */
            where += w;
            x->f_byteswritten += w;
            if (where == ebuf)
                break;
        }
        /* read any inputs that are ready */
        for (i = 0; i < cp_nproc; i++)
            if (i != cp_this
                && FD_ISSET((y = rtfifo_vec[i])->f_infd, &readset))
                rtfifo_getmore(y);
    }
    x->f_outmesshead = x->f_outmessbuf; /* empty the buffer */
#endif
}

/*  Send one long int.  It goes onto the output buffer of a
  rtfifo; if full the buffer is flushed.  If your C compiler
  doesn't do "inline" this could be replaced by a putc()-like macro. */

#ifdef PGCC
void rtfifo_sendlong(x, n)
#else
void rtfifo_sendlong(t_rtfifo *x, long int n)
#endif
{
    if (x->f_outmesshead == x->f_outmessend)
        rtfifo_flush(x);
    *(long *) (x->f_outmesshead) = n;
    x->f_outmesshead += sizeof(long);
}

void rtfifo_sendsym(t_rtfifo *x, t_symbol *s)   /* send a symbol */
{
    char *s1 = x->f_outmesshead, *es1 = x->f_outmessend, *s2 = s->s_name;
    if (s1 == es1)
        rtfifo_flush(x), s1 = x->f_outmesshead;
    do
    {
        *s1++ = *s2;
        if (s1 == es1)          /* if at end, rtfifflush and wrap */
        {
            x->f_outmesshead = s1;
            rtfifo_flush(x);
            s1 = x->f_outmessbuf;
        }
    }
    while (*s2++);
    s2 = s1;
    s1 += (-((long) s1)) & (sizeof(long) - 1);  /* forward to long boundary */
    while (s2 < s1)
        *(s2++) = 0;            /* zero out where forwarded over */
    /* one last chance to wrap */
    /* if (s1 == es1)
       {
       x->f_outmesshead = s1;
       rtfifo_flush(x);
       s1 = x->f_outmessbuf;
       } */
    x->f_outmesshead = s1;
}

/* send a message, to the object with host ID "id", on processor "cpno",
   with the given selector and arguments. */

void rtfifo_sendmess(int cpno, int what, long int id, t_symbol *s, int ac,
                     t_atom *av)
/* which CP receiver is on */
/* command (new, connect, etc.) */
/* ID of receiver */
/* selector of message to pass it */
/* argument count */
/* argument values */
{
    t_rtfifo *x;
    if (cpno < 0 || cpno >= cp_nproc || cpno == cp_this)
    {
        rtfifo_die("rtfifo_sendmess: cp number out of range");
        return;
    }
    x = rtfifo_vec[cpno];
    rtfifo_sendlong(x, (long) what);
    rtfifo_sendlong(x, id);
    if (s)
    {
        rtfifo_sendlong(x, (long) A_SYM);
        rtfifo_sendsym(x, s);
    }
    while (ac--)
    {
        if (av->a_type == A_SYM)
        {
            rtfifo_sendlong(x, (long) A_SYM);
            rtfifo_sendsym(x, av->a_w.w_sym);
        }
        else
        {
            rtfifo_sendlong(x, (long) (av->a_type));
            rtfifo_sendlong(x, av->a_w.w_long);
        }
        av++;
    }
    /* terminate it */
    rtfifo_sendlong(x, (long) A_NOTHING);
}

/* send a message to host.  Never called in the master CP.  */

rtfifo_sendtohost(long int *buf, int size)
/* size in bytes */
/* contents */
{
    t_rtfifo *x = rtfifo_vec[0];
    rtfifo_sendlong(x, (long) CP_SPECIAL);
    if (!cp_this)
        rtfifo_die("rtfifo_sendtohost");
    while (size > 0)
    {
        rtfifo_sendlong(x, *buf++);
        size -= sizeof(long);
    }
}

/* tell a CP which has sent a host update that the update has been sent
   out of master CP.  Called only from master. */

rtfifo_thanks(int cpno)
{
    t_rtfifo *x;
    if (cp_this || cpno <= 0 || cpno >= cp_nproc)
        rtfifo_die("rtfifo_thanks");
    x = rtfifo_vec[cpno];
    rtfifo_sendlong(x, (long) CP_SPECIAL);
    rtfifo_sendlong(x, (long) H_THANKS);
}

/* call a function remotely with one (word) argument */

rtfifo_fcall(int cpno, t_method fn, void *arg)
{
    t_rtfifo *x;
    if (cpno < 0 || cpno >= cp_nproc || cpno == cp_this)
        rtfifo_die("rtfifo_fcall");
    x = rtfifo_vec[cpno];
    rtfifo_sendlong(x, (long) CP_SPECIAL);
    rtfifo_sendlong(x, (long) H_FCALL);
    rtfifo_sendlong(x, (long) fn);
    rtfifo_sendlong(x, (long) arg);
}

/* do a function call on all CPs */

rtfifo_broadcast(t_method fn, long int arg)
{
    int i;
    (*fn) (arg);
    for (i = 0; i < cp_nproc; i++)
        if (i != cp_this)
            rtfifo_fcall(i, fn, (void *) arg);
}

/* mark the end of a tick.  For all rtfifos, this sends an end-of-tick packet. */
void rtfifo_tickout()
{
    int i;
    char *s1, *s2;
    for (i = 0; i < cp_nproc; i++)
        if (i != cp_this)
        {
            t_rtfifo *x = rtfifo_vec[i];
            rtfifo_sendlong(x, (long) CP_SPECIAL);
            rtfifo_sendlong(x, (long) H_ENDOFTICK);
            /* round up to signal boundary */
            s1 = x->f_outmesshead;
            s2 = s1 + ((-(long) s1) & (SIGALIGN - 1));
            while (s1 < s2)
                *(s1++) = 0;
            x->f_outmesshead = s2;
            /* if (x->f_outmesshead == x->f_outmessend) rtfifo_flush(x); */
            if (x->f_outmesshead != x->f_outmessbuf)
                rtfifo_flush(x);
            if (x->f_byteswritten & (SIGALIGN - 1))
                rtfifo_die("alignment");
        }
}

long rtfifo_getlong(t_rtfifo *x)        /*  Get a "long". */
{
    int onset = x->f_intail;
    if (onset & (sizeof(long) - 1))
        rtfifo_die("input buffer out of alignment");
    /* if the input channel is dead, just return zero. */
    while (x->f_inhead >= onset && x->f_inhead < onset + sizeof(long))
        if (rtfifo_getmore(x))
            return (0);
    x->f_intail = (onset >= INBUFSIZE - sizeof(long) ?
                   0 : onset + sizeof(long));
    /* {
       static int count = 0;
       if (count++ < 40)
       {
       int i;
       for (i = 0; i < 1000000; i++);
       rtfifo_debug("%d; get %lx\n", cp_this,
       *((long *)(x->f_inbuf + onset)));
       }
       } */
    return (*((long *) (x->f_inbuf + onset)));
}

t_symbol *rtfifo_getsym(t_rtfifo *x)    /*  Get a symbol. */
{
    char buf[MAXTRANSSYM];
    char *bp = buf, *ep = buf + MAXTRANSSYM;
    while (1)
    {
        *(long *) bp = rtfifo_getlong(x);
        if (!bp[0] || !bp[1] || !bp[2] || !bp[3])
            break;
        bp += 4;
        if (bp == ep)
        {
            bp[-1] = 0;
            break;
        }
    }
    return (gensym(buf));
}

/* get a host message from another CP and queue it on this one
   (which should be master.)  The constants CP_SPECIAL and H_QUEUE
   have already been taken off the input. */

void rtfifo_gethq(t_rtfifo *x)
{
    t_qmbuf qmbuf;
    int size;
    long *lp = (long *) qmbuf.b_s;
    qmbuf.b_m.m_command = H_QUEUE;
    size = (qmbuf.b_m.m_size = rtfifo_getlong(x)) - sizeof(qmbuf.b_q);
    qmbuf.b_q.qm_fn = (t_method) rtfifo_getlong(x);
    qmbuf.b_q.qm_ptr = (void *) rtfifo_getlong(x);
    if (size > HQ_MAX)
        rtfifo_die("overgrown HQ"), size = HQ_MAX;
    while (size > 0)
    {
        *lp++ = rtfifo_getlong(x);
        size -= sizeof(long);
    }
    exthq_setup(&qmbuf, x->f_cpno);
}

/* get a printout request from another CP and carry it out.  The constants
   CP_SPECIAL and H_PRINT have already been taken off the input. */

void rtfifo_getprintout(t_rtfifo *x)
{
    char buf[PR_MAX + 1];
    long *lp = (long *) buf;
    int size = rtfifo_getlong(x), s2 = size;
    if (size > PR_MAX)
        rtfifo_die("overgrown printout"), size = PR_MAX;
    while (size > 0)
    {
        *lp++ = rtfifo_getlong(x);
        size -= sizeof(long);
    }
    buf[s2] = 0;
    poststring(buf);
}

/* service a remote fcall request. */

void rtfifo_dofcall(t_rtfifo *x)
{
    t_method fn = (t_method) rtfifo_getlong(x);
    void *arg = (void *) rtfifo_getlong(x);
    (*fn) (arg);
}

/*  Get the incoming messages from the rtfifo "x" and handle them.  Here
  you can do blocking reads until you've got all the data.  Returns 1
  on EOF */

void rtfifo_infrom(t_rtfifo *x)
{
    int needbytes, foo = rtfifo_getlong(x);
    if (foo !=
        (0x12340000 | ((x->f_nexttick++) & 0xffff)) && !rtfifo_blowoff)
    {
#ifdef I860
        int i;
        LEDS[LED1] = 16;
        rtfifo_debug("out of sync");
        rtfifo_debug("x %lx", x);
        rtfifo_debug("inbuffer %lx", x->f_inbuf);
        rtfifo_debug("logicaltime %ld", logicaltime);
        rtfifo_debug("tick-count %lx", rtfifo_outtickcount);
        rtfifo_debug("next tick wanted %lx", x->f_nexttick - 1);
        rtfifo_debug("cp %ld", cp_this);
        rtfifo_debug("other cp %ld", x->f_cpno);
        rtfifo_debug("got value of %lx", foo);
        rtfifo_debug("in count: %lx", x->f_inhead);
        rtfifo_debug("phys in count: %lx", x->f_lctladdr->cpu_head);
        rtfifo_debug("in tail: %lx", x->f_intail);
        sys_spin(5);
#else
        rtfifo_die("out of sync\n");
#endif                          /* I860 */
    }
    while (1)
    {
        t_atom args[MAXMESSARGCOUNT];
        t_atom *argp = args, *earg = args + MAXMESSARGCOUNT;
        int ac = 0;
        long what;
        what = rtfifo_getlong(x);       /* get message type */
        if (what)               /* nonzero ID: ordinary message */
        {
            /* first arg is ID */
            argp->a_w.w_long = rtfifo_getlong(x);
            argp->a_type = A_LONG;
            argp++, ac++;
            while (1)
            {
                int type = rtfifo_getlong(x);
                if (!type)
                    break;
                argp->a_type = type;
                if (type == A_SYM)
                    argp->a_w.w_sym = rtfifo_getsym(x);
                else
                    argp->a_w.w_long = rtfifo_getlong(x);
                if (argp < earg)
                    argp++, ac++;
            }
            /* if (what < 0 || what >= CP_NEW + cp_nproc)
               rtfifo_die("garbage in\n"); */
            cp_dispatch((int) what, 0, ac, args);
        }
        else                    /* zero ID: special message */
        {
            int cmd = rtfifo_getlong(x);
            if (cmd == H_ENDOFTICK)
                break;
            else
                switch (cmd)
                {
                case H_PRINT:
                    rtfifo_getprintout(x);
                    break;
                case H_QUEUE:
                    rtfifo_gethq(x);
                    break;
                case H_THANKS:
                    hq_thanks();
                    break;
                case H_FCALL:
                    rtfifo_dofcall(x);
                    break;
                default:
                    rtfifo_die("%x: unknown inter-cp message", cmd);
                    break;
                }
        }
    }
    /* now go up to a SIGALIGN boundary */
    needbytes = (-x->f_intail) & (SIGALIGN - 1);
    while (x->f_inhead >= x->f_intail
           && x->f_inhead < x->f_intail + needbytes)
    {
        if (rtfifo_getmore(x))
            return;
    }
    x->f_intail += needbytes;
    if (x->f_intail > INBUFSIZE - sizeof(long))
        x->f_intail = 0;
}

/* do all input processing for the next tick.  Shuts down and returns
   nonzero if any other CP has exited */
rtfifo_tickin()
{
    while (rtfifo_deltaadvance > 0)
    {
        rtfifo_sigsout();
        rtfifo_donesigs();
        rtfifo_deltaadvance--;
    }
    for (rtfifo_cpfrom = 0; rtfifo_cpfrom < cp_nproc; rtfifo_cpfrom++)
        if (rtfifo_cpfrom != cp_this)
        {
            t_rtfifo *x = rtfifo_vec[rtfifo_cpfrom];
#if 0
            /* if schedule advance from other CP has been increased,  */
            if (x->f_countdown > 0)
            {
                x->f_countdown--;
            }
            else
            {
                rtfifo_infrom(x);
                if (rtfifo_blowoff)
                    return (1);
                while (x->f_countdown < 0)
                {
                    rtfifo_infrom(x);
                    if (rtfifo_blowoff)
                        return (1);
                    x->f_countdown++;
                }
            }
#else                           /* 0 */
            rtfifo_infrom(x);
            if (rtfifo_blowoff)
                return (1);
#endif                          /* 0 */
        }
    return (0);                 /* success */
}

void rtfifo_sigsin()
{
    t_sigio *s;
    t_rtfifo *x;
    int i;
    for (s = sig_inlist; s; s = s->s_next)
    {
        x = rtfifo_vec[s->s_cpno];
        /* rtfifo_die("insig\n"); */
        if (x->f_intail & (SIGALIGN - 1))
            rtfifo_die("in signal alignment");
        if (x->f_intail > INBUFSIZE - SIGSIZE)
        {
            /* rtfifo_die("\nwrap\n"); */
            while (x->f_inhead >= x->f_intail)
            {
                if (rtfifo_getmore(x))
                    return;
            }
            x->f_intail = 0;
        }
        while (x->f_inhead >= x->f_intail &&
               x->f_inhead < x->f_intail + SIGSIZE)
            rtfifo_getmore(x);
        *(s->s_where) = (float *) (x->f_inbuf + x->f_intail);
        x->f_intail += SIGSIZE;
        if (x->f_intail >= INBUFSIZE)
            x->f_intail = 0;
    }

}

void rtfifo_sigsout()
{
    t_sigio *s;
    t_rtfifo *x;
    int i;

    /* output signals go into the output buffer -- problem:
       this limits their number.  Care is taken that the signal
       won't wrap around in the receiver's input buffer. */

    rtfifo_tickout();           /* terminate message output and flush FIFOs */
    for (s = sig_outlist; s; s = s->s_next)
    {
        int phase;
        x = rtfifo_vec[s->s_cpno];
        phase = (x->f_byteswritten + (x->f_outmesshead - x->f_outmessbuf))
            & (INBUFSIZE - 1);
        if (phase > (INBUFSIZE - SIGSIZE))
        {
            while (phase < INBUFSIZE)
            {
                *(long *) (x->f_outmesshead) = 0;
                x->f_outmesshead += sizeof(long);
                phase += sizeof(long);
            }
            phase = 0;
        }
        *(s->s_where) = (float *) (x->f_outmesshead);
        x->f_outmesshead += SIGSIZE;
        if (x->f_outmesshead > x->f_outmessend)
            rtfifo_die("interprocessor signal overflow\n");
    }

    /* if there's no DSP running we'd better zero out the sound
       buffers or else they'll be trash and the receiving CP might
       start up before it starts getting real signals from this CP. */

    if (!u_chain)
    {
        for (s = sig_outlist; s; s = s->s_next)
            cu_zero(*(s->s_where), IPCVS);
    }
    rtfifo_sigsused = 1;

    /* special marker to detect out-of-sync.  This marker should be
       after all the signals, but before any control messages are written.
       Since control mesages can come from the "print~" ugen we put this here
       and not at the beginning of rtfifo_donesigs, as you'd think it should
       be.  It can crash FTS if DSP puts enough info at the end of the outmessbuf
       to force a flush.  This overflow condition should somehow be detected...
     */
    {
        int rats;
        for (rats = 0; rats < cp_nproc; rats++)
            if (rats != cp_this)
                rtfifo_sendlong(rtfifo_vec[rats],
                                0x12340000L | (rtfifo_outtickcount &
                                               0xffff));
        rtfifo_outtickcount++;
    }
}

/* set up pointers to input and output signal locations for this tick. */
void rtfifo_setupsigs()
{
    rtfifo_sigsin();
    if (rtfifo_deltaadvance >= 0)
        rtfifo_sigsout();
}

/* called after the input signals are used and the outputs are ready to send
   out.  */

rtfifo_donesigs()
{
    t_rtfifo **x;
    int i;
    rtfifo_sigsused = 0;

    if (rtfifo_deltaadvance < 0)
    {
        rtfifo_deltaadvance++;
        return;
    }
#ifdef I860
    io_waitfordma();
    for (i = 0, x = rtfifo_vec; i < cp_nproc; x++, i++)
        if (i != cp_this)
            (*x)->f_rctladdr->cpu_head = (*x)->f_remotehead;
#endif                          /* I860 */
}

#if 0
/* reset scheduler advance for cp number i. */
rtfifo_setadvancefor(int i, int adv)
{
    int delta;
    t_rtfifo *x;
    if (i == cp_this)
        return;
    if (i < 0 || i >= cp_nproc || adv < MINADVANCE || adv > MAXADVANCE)
        bug("rtfifo_setadvancefor");
    x = rtfifo_vec[i];
    delta = adv - x->f_inadvance;
    x->f_countdown += delta;
    x->f_inadvance = adv;
}
#endif

/* reset scheduler advance */
void rtfifo_setadvance(newadvance)
{
    rtfifo_deltaadvance += (newadvance - rtfifo_advance);
    rtfifo_advance = newadvance;
}

rtfifo_print()
{
    long count[MAXCP];
    int i;
    t_sigio *s;
    for (i = 0; i < cp_nproc; i++)
        count[i] = 0;
    for (s = sig_outlist; s; s = s->s_next)
        count[s->s_cpno]++;
    post("IPC sigs: ");
    for (i = 0; i < cp_nproc; i++)
        postlong(count[i]);
}

/* In simulation, the output filedesc should
   have been set to F_ASYNC via fcntl, so that a write to it will
   only write as many bytes at a time as the pipe can take, rather than
   waiting for the whole thing to go through.
   In realtime the filedescs have different meanings:
   infd: slot number of this CP;
   outfd: slot number of the other CP.
   */

void rtfifo_init(int cpno, int advance, int infd, int outfd, int nodma)
/* ID number of other CP */
/* scheduler advance in ticks */
/* filedesc for input */
/* filedesc for output */
/* inhibit DMA -- I860 only */
{
    t_rtfifo *x;
    /* sa_printf("init %d %d %d %d %d\n",
       cpno, advance, infd, outfd, nodma); */
    if (cpno < 0 || cpno >= cp_nproc)
        rtfifo_die("rtfifo: bad CP ID number");
    x = (t_rtfifo *) getbytes(sizeof(t_rtfifo));
    if (!x)
        hate("rtfifo_init: out of space");
    if (!(x->f_outmessbuf = malloc(OUTBUFSIZE + SIGALIGN)))
        hate("out of space");
    x->f_outmessbuf += (-(long) x->f_outmessbuf) & (SIGALIGN - 1);

#ifdef I860
    {
        char *slot = sys_slottoaddr(outfd);
        int i;
        /* don't use DMA if it's your neighbor (or if DMA inhibited) */
        if (nodma || outfd == (cp_slotno ^ 1))
            x->f_dmathresh = 0x7fffffff;
        else
            x->f_dmathresh = 256;
        x->f_remotehead = 0;
        x->f_lctladdr = &PROTO->p_ftscpu[outfd];
        x->f_rctladdr =
            (t_ftscpu *) (slot +
                          (0x1ffffff & (int) (&PROTO->p_ftscpu[infd])));
        x->f_rbufaddr =
            slot + (FIFOADDR -
                    INBUFSIZE * (cp_this > cpno ? cp_this : cp_this + 1));
        x->f_inbuf =
            (char *) (LDRAM_CDI + FIFOADDR -
                      INBUFSIZE * (cpno > cp_this ? cpno : cpno + 1));
        for (i = 0; i < 65536 / 4; i++)
            ((long *) (x->f_inbuf))[i] = 0x87654321;

        /*
           rtfifo_debug("thresh %x", x->f_dmathresh);
           rtfifo_debug("cpu %lx", cpno);
           rtfifo_debug("ctl in %lx", x->f_lctladdr);
           rtfifo_debug("buf in %lx", x->f_inbuf);
           rtfifo_debug("ctl out %lx", x->f_rctladdr);
           rtfifo_debug("buf out %lx\n", x->f_rbufaddr);
           rtfifo_debug("flags %lx\n", &(PROTO->p_ftsflags));
         */
    }
#else
    {
        char sbuf[80];
        sprintf(sbuf, "logfile-%d-%d", cp_this, cpno);
#ifdef MAKELOGFILE              /* unused; was useful while testing the protocol */
        if ((x->f_logfd = creat(sbuf, 0666)) < 0)
        {
            perror("creat");
            hate("couldn\'t open logfile");
        }
#else
        x->f_logfd = -1;
#endif                          /* MAKELOGFILE */
        if (!(x->f_inbuf = malloc(INBUFSIZE + SIGALIGN)))
            hate("out of space");
        x->f_inbuf += (-(long) x->f_inbuf) & (SIGALIGN - 1);
        x->f_infd = infd;
        x->f_outfd = outfd;
    }
#endif                          /* I860 */
    x->f_inhead = x->f_intail = 0;
    x->f_cpno = cpno;
    x->f_outmesshead = x->f_outmessbuf;
    x->f_outmessend = x->f_outmessbuf + OUTBUFSIZE;

    if (advance < 1)
    {
        error("scheduler advance for CP %d too small; set to 1", advance);
        advance = 1;
    }
    else if (advance > MAXADVANCE)
    {
        error("scheduler advance for CP %d too large; set to %d",
              cpno, MAXADVANCE);
        advance = MAXADVANCE;
    }
    x->f_inadvance = advance;
    x->f_nexttick = 0;
    x->f_byteswritten = 0;
    x->f_insigcount = x->f_outsigcount = 0;
    rtfifo_vec[cpno] = x;
    rtfifo_sendlong(x, 0x12340000L);
}
