/*
 * copyright 1991-1994 IRCAM,  1995-1996 UCSD.
 * 
 * The server for asynchronous disk I/O as used in d_disk.c.  This is
 * not part of the Max/FTS program; rather; it is compiled separately
 * into the "max-sounddisk" program.
 * 
 * For the ISPW, max-sounddisk runs on the NeXT host.
 * The host serves FTS for a number of asynchronous sound paths
 * to and from disk.  the Max program spawns max-sounddisk processes.
 * The FTS resident memory is found via the PROTO macro,  which is
 * defined in ../6a/proto56.h
 *
 * For the SGI, one max-sounddisk process is spawned by FTS for each
 * instance of readsf, writesf,  or vreadsf.  Here "PROTO" points
 * to the beginning of shared memory  
 */

#ifdef NEXTSTEP
#include <sound/soundstruct.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include "../6a/6a.h"
#include "../6a/proto56.h"
extern int errno;
#endif

#ifdef __unix
#include "cp_sound.h"
#include "d_disk.h"
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
long lseek();
#define MEM_PAGESIZE 4096

#endif

SNDSoundStruct io_soundheader =
    { SND_MAGIC, sizeof(SNDSoundStruct), 0x10000000,
    SND_FORMAT_LINEAR_16, 44100., 1, 0x62617a00
};

#define FLOATSAMPLESIZE (sizeof(float))
#define SHORTSAMPLESIZE (sizeof(short))

#define F_READ 1
#define F_WRITE 2
#define F_BIDIRECTIONAL 3

/* The server's per-open-filedesc state is described here. */

typedef struct _hasyncfd {
    int h_fd;                   /* UNIX filedesc */
    long h_hostonset;           /* host onset for FIFO on CP */
    long h_hostbackonset;       /* host onset for going backwards */
    long h_size;                /* size of FIFO on CP */
    long h_buf;                 /* FIFO buffer on CP */
    char h_function;            /* read/write flag */
    char h_first;               /* first-transfer flag */
    long h_chans;               /* number of audio channels */
    long h_sr;                  /* sample rate */
    long h_skipbytes;           /* bytes to drop on first read */
    long h_soundbegin;          /* onset of beginning of sound */
    long h_soundend;            /* onset of end of sound (F_BIDIRECTIONAL only) */
    long h_fileposforhead;      /* position of file corresponding to CP head */
} t_hasyncfd;

#ifdef NEXTSTEP
typedef struct hvec             /* there are MAXAFD of these per CP */
{
    t_hasyncfd h[MAXAFD];
} t_hvec;

t_hvec sd_asyncfds[MAXSLOTNO];  /* ... and up to MAXSLOTNO CPs */
#endif

#ifdef __unix
/*
 * in Unix we just read straight into FTS's memory.
 */
static void *sd_shmaddr;        /* Shared memory address */
static int sd_shmid;            /* Shared memory ID */
static int sd_memoffset;        /* add to FTS pointers to get shmem pointers */
static t_asyncfd *sd_asyncfd;   /* FTS's afd structure in FTS mem map */
t_hasyncfd sd_hasyncfd;         /* server's corresponding structure *

                                   /* mmap_read, etc, are just memory copies: */
#define mmap_readlong(x) (*(long *)(((char *)(x)) + sd_memoffset))
#define mmap_readbyte(x) (*(char *)(((char *)(x)) + sd_memoffset))
#define mmap_writelong(y,  x) (*(long *)(((char *)(x)) + sd_memoffset) = (y))
#define mmap_writebyte(y,  x) (*(char *)(((char *)(x)) + sd_memoffset) = (y))
#define mmap_read(local, remote, size) \
    (bcopy((void *)(((char *)(remote)) + sd_memoffset),\
  (void *)(local),  (size_t)(size)))
#define mmap_write(local, remote, size) \
    (bcopy((void *)(local),\
  (void *)(((char *)(remote)) + sd_memoffset),  (size_t)(size)))
  /* there's no notion of "slot" here. */
#endif

#ifdef NEXTSTEP
/*
 * on the NeXT we buffer everything on the host and copy it down
 * via mmap_write or up via mmap_read.
 */
char *sd_soundbuf;              /* the local buffer for sound */
int sd_slots[MAXSLOTNO];        /* true if that slot is used */
#endif

#define MINASYNCSIZE 0x10000    /* smallest amount we're willing to read */
#define MAXASYNCSIZE 0x30000    /* most we'll read at once */

int sd_serverno;                /* which of several possible servers we are */
int sd_open;                    /* true if last pass revealed that an FD is open */
int sd_debug = 0;

int sd_getinfo(int fd, long *headersize, int *nchans, float *srate,
               int *bigendian, int *floating)
{
    SNDSoundStruct header;
    if (read(fd, &header, sizeof(header)) < sizeof(header))
        return (-1);
    if (header.magic != SND_MAGIC)
    {
        *headersize = 1024;
        *nchans = 1;
        *srate = 44100.;
        *bigendian = 1;
        *floating = 0;
    }
    else
    {
        *headersize = header.dataLocation;
        *nchans = header.channelCount;
        *srate = header.samplingRate;
        *bigendian = 1;
        *floating = (header.dataFormat == SND_FORMAT_FLOAT);
    }
}

static int mod(x, y)
{
    if (x >= 0)
        return (x % y);
    else
        return ((y - ((-x) % y)) % y);
}

int sd_getdiskfor(t_hasyncfd *hafd, long fileaddr, long transfersize,
                  char *soundbuf)
{
    int realfileaddr, readlength, zerostart, zeroend, readtowhere;
    if (transfersize < 0)
        return (-1);
    if (fileaddr + transfersize < hafd->h_soundbegin ||
        fileaddr > hafd->h_soundend)
    {
        bzero(soundbuf, transfersize);
        return (0);
    }
    readtowhere = (fileaddr >= 0 ? 0 : -fileaddr);
    zerostart = (fileaddr > hafd->h_soundbegin ?
                 0 : hafd->h_soundbegin - fileaddr);
    zeroend = (fileaddr + transfersize - hafd->h_soundend > 0 ?
               fileaddr + transfersize - hafd->h_soundend : 0);
    realfileaddr = (fileaddr < 0 ? 0 : fileaddr);
    readlength = transfersize - (realfileaddr - fileaddr) - zeroend;
    if (sd_debug)
        fprintf(stderr, "sd_getdiskfor addr %d, length %d\n",
                fileaddr, transfersize);
    if (sd_debug)
        fprintf(stderr, "readtowhere %d, readlength %d, zero %d %d\n",
                readtowhere, readlength, zerostart, zeroend);
    if (sd_debug)
        fprintf(stderr, "realfileaddr %d\n", realfileaddr);
    if (readlength < 0 || readtowhere + readlength > transfersize ||
        readtowhere < 0 || zerostart < 0 || zerostart > transfersize ||
        zeroend < 0 || zeroend > transfersize)
    {
        fprintf(stderr, "bug 3\n");
        return (-1);
    }
    if (lseek(hafd->h_fd, realfileaddr, 0) < 0)
        return (-1);
    if (read(hafd->h_fd, soundbuf + readtowhere, readlength) < readlength)
        return (-1);
    if (zerostart)
        bzero(soundbuf, zerostart);
    if (zeroend)
        bzero(soundbuf + transfersize - zeroend, zeroend);
    return (0);
}

sd_init()
{
    int i, j;
#ifdef NEXTSTEP
    for (i = 0; i < MAXSLOTNO; i++)
        for (j = 0; j < MAXAFD; j++)
            sd_asyncfds[i].h[j].h_fd = -1;
    if (!(sd_soundbuf = (char *) malloc(MAXASYNCSIZE)))
    {
        fprintf(stderr, "no space for sound buffer\n");
        exit(1);
    }
#endif
#ifdef __unix
    sd_hasyncfd.h_fd = -1;
#endif
}

static fd_set readset, writeset, exceptset;

sd_minisleep(msec)              /* wait < 1 sec */
{
    struct timeval timout;
    if (sd_debug)
        fprintf(stderr, "minisleep %d\n", msec);
    timout.tv_sec = msec / 1000;
    timout.tv_usec = (msec % 1000) * 1000;
    FD_ZERO(&writeset);
    FD_ZERO(&readset);
    FD_ZERO(&exceptset);
    select(0, &readset, &writeset, &exceptset, &timout);
}

int sd_stdinready()             /* return 1 if standard in "selects" */
{
    struct timeval timout;
    timout.tv_sec = 0;
    timout.tv_usec = 0;
    FD_ZERO(&writeset);
    FD_ZERO(&readset);
    FD_ZERO(&exceptset);
    FD_SET(0, &readset);
    FD_SET(0, &exceptset);
    if (select(1, &readset, &writeset, &exceptset, &timout) < 0
        || FD_ISSET(0, &readset) || FD_ISSET(0, &exceptset))
        return (1);
    else
        return (0);
}

#define BLOCKSIZEGUESS 8192
static int didsomething;

    /* poll one async disk connection */

void sd_dopoll(t_hasyncfd *hafd, t_asyncfd *cpfd)
{
    t_afdhead hostcopy;
    char zero = 0;
    /* pick up the header */
    mmap_read(&hostcopy, &cpfd->afd_h, sizeof(hostcopy));
    /* fprintf(stderr, "%lx %d %d %d\n", &cpfd->afd_h,
       (int)(hostcopy.h_openrequest),
       (int)(hostcopy.h_closerequest),
       (int)(hostcopy.h_eof)); */
    /* ignore if serverno doesn't match us */
    if (hostcopy.h_serverno != sd_serverno)
    {
        if (hafd->h_fd >= 0)
        {
            close(hafd->h_fd);
            hafd->h_fd = -1;
        }
        return;
    }
    /* if it's a write-to-disk channel, service close request
       only after host has transferred all the data  */
    if (hostcopy.h_closerequest &&
        ((hafd->h_function != F_WRITE) ||
         hafd->h_hostonset == mmap_readlong(&cpfd->afd_cponset)))
    {
        if (hafd->h_fd >= 0)
        {
            close(hafd->h_fd);
            hafd->h_fd = -1;
        }
        mmap_writebyte(zero, &cpfd->afd_h.h_closerequest);
        didsomething = 1;
    }
    if (hostcopy.h_openrequest) /* open for reading or writing */
    {
        char name[AFD_NAMESIZE], *cp, cpres;
        long cpfilename;
        int result, writerequest, bidirectional;
        errno = 0;
        if (hafd->h_fd >= 0)
        {
            if (hafd->h_function == F_WRITE)
                fprintf(stderr,
                        "max-sounddisk: bug:open to write before close finished");
            close(hafd->h_fd);
            hafd->h_fd = -1;
        }
        /* fprintf(stderr, "name addr at %x\n", &cpfd->afd_filename); */
        cpfilename = mmap_readlong(&cpfd->afd_filename);
        cp = name;
        do
        {
            *cp = mmap_readbyte(cpfilename++);
        }
        while (*cp++);
        if (writerequest = (hostcopy.h_openrequest == AFDREQ_OPENTOWRITE))
            result = creat(name, 0666);
        else
            result = open(name, 0);
        if (sd_debug)
            fprintf(stderr, "dopoll 5\n");
        if (result >= 0)
        {
            if (writerequest)
            {
                hafd->h_chans = mmap_readlong(&cpfd->afd_nchans);
                hafd->h_sr = mmap_readlong(&cpfd->afd_sr);
            }
            else
            {
                long skipbytes;
                long headersize;
                int nchans, bigendian, floating;
                float srate;
                long onset;
                char onsetformat;
                onset = mmap_readlong(&cpfd->afd_onset);
                onsetformat = mmap_readbyte(&cpfd->afd_onsetformat);
                bidirectional = mmap_readbyte(&cpfd->afd_bidirectional);
                if (sd_getinfo(result, &headersize, &nchans, &srate,
                               &bigendian, &floating) < 0)
                {
                    result = -1;
                    goto donelooking;
                }
                hafd->h_soundbegin = headersize;
                if (bidirectional)
                {
                    if (sd_debug)
                        fprintf(stderr, "bidirectional\n");
                    hafd->h_soundend = lseek(result, 0L, 2);
                    if (hafd->h_soundend < 0 || lseek(result, 0L, 0) < 0)
                    {
                        result = -1;
                        goto donelooking;
                    }
                }
                if (onsetformat == F_MSEC)
                {
                    skipbytes = headersize +
                        (nchans *
                         (floating ? FLOATSAMPLESIZE : SHORTSAMPLESIZE)) *
                        (int) (srate * .001 * onset);
                }
                else
                    skipbytes = onset;
                if (skipbytes < 0)
                    skipbytes = 0;
                skipbytes &= (~3);
                hafd->h_fileposforhead = skipbytes;
                if ((!bidirectional) && skipbytes > BLOCKSIZEGUESS)
                {
                    long seeker = skipbytes & (~(BLOCKSIZEGUESS - 1));
                    if (lseek(result, seeker, 0) < 0)
                    {
                        result = -1;
                        goto donelooking;
                    }
                    skipbytes -= seeker;
                }
                hafd->h_skipbytes = skipbytes;
                mmap_writelong((long) nchans, &cpfd->afd_nchans);
                mmap_writelong((long) srate, &cpfd->afd_sr);
            }
        }
        if (sd_debug)
            fprintf(stderr, "dopoll 7\n");
        hafd->h_size = mmap_readlong(&cpfd->afd_size);
        hafd->h_buf = mmap_readlong(&cpfd->afd_buf);
        if (hafd->h_size <= MAXASYNCSIZE ||
            (bidirectional && hafd->h_size <= 2 * MAXASYNCSIZE))
        {
            fprintf(stderr, "maxsounddisk: buffer too small\n");
            result = -1;
        }
        if (sd_debug)
            fprintf(stderr, "buf %lx, size %ld\n",
                    hafd->h_buf, hafd->h_size);
      donelooking:
        if (result < 0)
        {
            cpres = errno;
            if (!cpres)
                cpres = 0xff;
            hafd->h_fd = -1;
        }
        else
        {
            cpres = 0;
            hafd->h_fd = result;
        }
        hafd->h_hostonset = 0;
        if (sd_debug)
            fprintf(stderr, "set hostonset (c) %ld\n", hafd->h_hostonset);
        hafd->h_first = 1;
        hafd->h_function = (writerequest ? F_WRITE :
                            (bidirectional ? F_BIDIRECTIONAL : F_READ));
        mmap_writelong(result, &cpfd->afd_openresult);
        mmap_writelong(0L, &cpfd->afd_hostonset);
        mmap_writebyte(cpres, &cpfd->afd_h.h_eof);
        mmap_writebyte(zero, &cpfd->afd_h.h_openrequest);
        didsomething = 1;
        if (sd_debug)
            fprintf(stderr, "%s: fd = %d; write = %d\n",
                    name, hafd->h_fd, writerequest);
    }
    if (hafd->h_fd >= 0)
    {
        long *lp1, *lp2, count, byteswanted, transfersize;
        int ret;
        sd_open = 1;
        errno = 0;
        if (hafd->h_function == F_WRITE)        /* write to disk */
        {
            long tail = hafd->h_hostonset;
            long head = mmap_readlong(&cpfd->afd_cponset);
            if (sd_debug)
                fprintf(stderr, "write: head %ld, tail %ld\n", head, tail);
            byteswanted = head - tail;
            if (byteswanted < 0)
                byteswanted += hafd->h_size;
            if (!hostcopy.h_closerequest)
            {
                byteswanted &= (~(MEM_PAGESIZE - 1));
                byteswanted -= MEM_PAGESIZE;
                if (byteswanted < MINASYNCSIZE)
                    goto skipit;
            }
            transfersize = (byteswanted > MAXASYNCSIZE ?
                            MAXASYNCSIZE : byteswanted);
            if (sd_debug)
                fprintf(stderr, "xfer: %ld\n", transfersize);
            if (tail + transfersize > hafd->h_size)
                transfersize = hafd->h_size - tail;
#ifdef NEXTSTEP
            mmap_read((long *) sd_soundbuf, hafd->h_buf + tail,
                      transfersize);
            if (hafd->h_first)
            {
                io_soundheader.channelCount = hafd->h_chans;
                io_soundheader.samplingRate = hafd->h_sr;
                *(SNDSoundStruct *) sd_soundbuf = io_soundheader;
                hafd->h_first = 0;
            }
            ret = write(hafd->h_fd, sd_soundbuf, transfersize);
#endif
#ifdef __unix
            if (hafd->h_first)
            {
                io_soundheader.channelCount = hafd->h_chans;
                io_soundheader.samplingRate = hafd->h_sr;
                mmap_write(&io_soundheader, hafd->h_buf,
                           sizeof(io_soundheader));
                hafd->h_first = 0;
            }
            ret = write(hafd->h_fd, hafd->h_buf + tail + sd_memoffset,
                        transfersize);
#endif
            if (ret < transfersize)
            {
                if (!(errno & 0xff))
                    errno = 0xff;
                mmap_writebyte((char) errno, &cpfd->afd_h.h_eof);
                perror("sound file write");
                ret = 0;
                close(hafd->h_fd);
                hafd->h_fd = -1;
            }
            tail += ret;
            if (tail == hafd->h_size)
                tail = 0;
            mmap_writelong(tail, &cpfd->afd_hostonset);
            hafd->h_hostonset = tail;
            if (sd_debug)
                fprintf(stderr, "set hostonset to %ld\n",
                        hafd->h_hostonset);
            /* fprintf(stderr,
               "after: head %ld, tail %ld\n", head, tail); */
            didsomething = 1;
          skipit:;
        }
        else if (hafd->h_function == F_READ)    /* read from disk */
        {
            long head = hafd->h_hostonset;
            long tail = mmap_readlong(&cpfd->afd_cponset);
            byteswanted = tail - head;
            if (byteswanted <= 0)
                byteswanted += hafd->h_size;
            byteswanted--;
            if (sd_debug)
                fprintf(stderr,
                        "read: head %ld, tail %ld, wanted %lx\n", head,
                        tail, byteswanted);
            if (byteswanted > MINASYNCSIZE)
            {
                long cpxfer1, cpxfer2, cpxfertot;
                transfersize = byteswanted + hafd->h_skipbytes;
                transfersize = (transfersize > MAXASYNCSIZE ?
                                MAXASYNCSIZE : transfersize);
                transfersize &= (~(BLOCKSIZEGUESS - 1));
#ifdef NEXTSTEP
                ret = read(hafd->h_fd, sd_soundbuf, transfersize);
                if (ret < 0)
                {
                    if (!(errno & 0xff))
                        errno = 0xff;
                    mmap_writebyte((char) errno, &cpfd->afd_h.h_eof);
                    perror("sound file read");
                    close(hafd->h_fd);
                    hafd->h_fd = -1;
                    return (1);
                }
                if (sd_debug)
                    fprintf(stderr,
                            "do read: %ld, ret %ld, skip %ld\n",
                            transfersize, ret, hafd->h_skipbytes);
                ret &= (~3);
                cpxfertot = ret - hafd->h_skipbytes;
                cpxfer1 = (cpxfertot > hafd->h_size - head ?
                           hafd->h_size - head : cpxfertot);
                cpxfer2 = cpxfertot - cpxfer1;
                if (sd_debug)
                    fprintf(stderr, "x1 %ld, x2 %ld\n", cpxfer1, cpxfer2);
                if (cpxfer1)
                {
                    mmap_write(sd_soundbuf + hafd->h_skipbytes,
                               hafd->h_buf + head, cpxfer1);
                    head += cpxfer1;
                }
                if (cpxfer2)
                {
                    mmap_write(sd_soundbuf + hafd->h_skipbytes + cpxfer1,
                               hafd->h_buf, cpxfer2);
                    head = cpxfer2;
                }
#endif
#ifdef __unix
                if (head + transfersize > hafd->h_size)
                    transfersize = hafd->h_size - head;
                /* {
                   static char buf[0x40000];
                   ret = read(hafd->h_fd, buf, transfersize);
                   if (ret > 0)
                   bcopy(buf,  hafd->h_buf + head + sd_memoffset,  ret);
                   } */
                ret = read(hafd->h_fd, hafd->h_buf + head + sd_memoffset,
                           transfersize);
                if (sd_debug)
                    fprintf(stderr,
                            "do read: %ld, ret %ld\n", transfersize, ret);
                if (ret < 0)
                {
                    if (!(errno & 0xff))
                        errno = 0xff;
                    mmap_writebyte((char) errno, &cpfd->afd_h.h_eof);
                    perror("sound file read");
                    close(hafd->h_fd);
                    hafd->h_fd = -1;
                    return;
                }
                head += ret;
#endif
                if (head == hafd->h_size)
                    head = 0;
                mmap_writelong(head, &cpfd->afd_hostonset);
                hafd->h_hostonset = head;
                if (sd_debug)
                    fprintf(stderr, "set hostonset (b) %ld\n",
                            hafd->h_hostonset);
                hafd->h_skipbytes = 0;
                if (ret < transfersize)
                {
                    close(hafd->h_fd);
                    hafd->h_fd = -1;
                    mmap_writebyte(0xff, &cpfd->afd_h.h_eof);
                }
                didsomething = 1;
                if (sd_debug)
                    fprintf(stderr,
                            "after: head %ld, tail %ld\n", head, tail);
            }
        }
        else                    /* read -- bidirectional */
        {
#ifdef __unix
            static char *sd_soundbuf;
            if (!sd_soundbuf)
            {
                if (!(sd_soundbuf = (char *) malloc(MAXASYNCSIZE)))
                    fprintf(stderr, "sounddisk: malloc failed\n");
            }
#endif
            if (hafd->h_first)
            {
                /* the first read, in which the forward and backward
                   fill pointers start at zero.  The first read points
                   forward except that the seek is rounded down to a
                   multiple of BLOCKSIZEGUESS so that _skipbytes_ bytes
                   are written behind the starting position. */

                long skipbytes = hafd->h_fileposforhead &
                    (BLOCKSIZEGUESS - 1);
                ret = sd_getdiskfor(hafd,
                                    hafd->h_fileposforhead - skipbytes,
                                    (long) MAXASYNCSIZE, sd_soundbuf);
                if (sd_debug)
                    fprintf(stderr,
                            "first varispeed %ld (skip %ld) -> %d\n",
                            hafd->h_fileposforhead - skipbytes,
                            skipbytes, ret);
                if (ret < 0)
                {
                    if (!(errno & 0xff))
                        errno = 0xff;
                    mmap_writebyte((char) errno, &cpfd->afd_h.h_eof);
                    perror("sound file read");
                    close(hafd->h_fd);
                    hafd->h_fd = -1;
                    return;
                }
                mmap_write(sd_soundbuf + skipbytes,
                           hafd->h_buf, MAXASYNCSIZE - skipbytes);
                if (skipbytes)
                    mmap_write(sd_soundbuf,
                               hafd->h_buf + hafd->h_size - skipbytes,
                               skipbytes);
                hafd->h_first = 0;
                hafd->h_hostonset = MAXASYNCSIZE - skipbytes;
                hafd->h_hostbackonset = mod(hafd->h_size - skipbytes,
                                            hafd->h_size);
                hafd->h_fileposforhead += hafd->h_hostonset;
                didsomething = 1;
                if (sd_debug)
                    fprintf(stderr, "did it\n");
            }
            else
            {
                /* not the first read.  Decide whether forward or 
                   backward is more urgent, and whether the more urgent
                   one can take MINASYNCSIZE bytes yet.  */

                long head, backhead, tail, fwdbyteswanted, backbyteswanted,
                    diff, diff2, mostwanted;
                head = hafd->h_hostonset;
                backhead = hafd->h_hostbackonset;
                tail = mmap_readlong(&cpfd->afd_cponset);

                /* compute forward bytes wanted. */
                diff = head - tail;
                if (diff < 0)
                    diff += hafd->h_size;
                fwdbyteswanted = hafd->h_size / 2 - diff;

                /* the head shouldn't be more than a half buffer
                   past the tail. */

                if (fwdbyteswanted < 0)
                    fwdbyteswanted = 0;

                /* compute backward bytes wanted. */
                diff = tail - backhead;
                if (diff < 0)
                    diff += hafd->h_size;
                backbyteswanted = hafd->h_size / 2 - diff;
                if (backbyteswanted < 0)
                    backbyteswanted = 0;

                if (sd_debug)
                    fprintf(stderr,
                            "vread: head %ld, backhead %ld, tail %ld\n",
                            head, backhead, tail);
                if (sd_debug)
                    fprintf(stderr,
                            "fwdbyteswanted %ld, backbyteswanted %ld\n",
                            fwdbyteswanted, backbyteswanted);

                mostwanted = (fwdbyteswanted > backbyteswanted ?
                              fwdbyteswanted : backbyteswanted);
                if (mostwanted > MINASYNCSIZE)  /* do one */
                {
                    long cpxfer1, cpxfer2, cpxfertot, fileaddr, bufphase,
                        newhead, newbackhead;
                    transfersize = (mostwanted > MAXASYNCSIZE ?
                                    MAXASYNCSIZE : mostwanted);
                    transfersize &= (~(BLOCKSIZEGUESS - 1));
                    if (fwdbyteswanted > backbyteswanted)
                    {
                        /* read forward */
                        fileaddr = hafd->h_fileposforhead;
                        newhead = mod(head + transfersize, hafd->h_size);
                        diff = mod(backhead - head, hafd->h_size);
                        diff2 = mod(backhead - newhead, hafd->h_size);
                        if (diff < hafd->h_size / 2 &&
                            diff2 > hafd->h_size / 2)
                            newbackhead = newhead;
                        else
                            newbackhead = backhead;
                        hafd->h_fileposforhead += transfersize;
                        bufphase = head;
                    }
                    else
                    {
                        fileaddr = hafd->h_fileposforhead -
                            mod(head - tail, hafd->h_size) -
                            mod(tail - backhead, hafd->h_size) -
                            transfersize;
                        newbackhead = mod(backhead - transfersize,
                                          hafd->h_size);
                        diff = mod(backhead - head, hafd->h_size);
                        diff2 = mod(newbackhead - head, hafd->h_size);
                        if (diff < hafd->h_size / 2 &&
                            diff2 > hafd->h_size / 2)
                            newhead = newbackhead;
                        else
                            newhead = head;
                        hafd->h_fileposforhead -= mod(head - newhead,
                                                      hafd->h_size);
                        bufphase = newbackhead;
                    }

                    ret = sd_getdiskfor(hafd, fileaddr, transfersize,
                                        sd_soundbuf);
                    if (sd_debug)
                        fprintf(stderr,
                                "varispeed %ld (size %ld) -> %d\n",
                                fileaddr, transfersize, ret);
                    if (ret < 0)
                    {
                        if (!(errno & 0xff))
                            errno = 0xff;
                        mmap_writebyte((char) errno, &cpfd->afd_h.h_eof);
                        perror("sound file read");
                        close(hafd->h_fd);
                        hafd->h_fd = -1;
                        return;
                    }
                    cpxfer1 = (transfersize > hafd->h_size - bufphase ?
                               hafd->h_size - bufphase : transfersize);
                    cpxfer2 = transfersize - cpxfer1;
                    if (sd_debug)
                        fprintf(stderr,
                                "x1 %ld, x2 %ld bufphase %ld\n", cpxfer1,
                                cpxfer2, bufphase);
                    if (cpxfer1)
                        mmap_write(sd_soundbuf, hafd->h_buf + bufphase,
                                   cpxfer1);
                    if (cpxfer2)
                        mmap_write(sd_soundbuf + cpxfer1, hafd->h_buf,
                                   cpxfer2);

                    mmap_writelong(hafd->h_hostonset = newhead,
                                   &cpfd->afd_hostonset);
                    mmap_writelong(hafd->h_hostbackonset = newbackhead,
                                   &cpfd->afd_hostbackonset);
                    if (sd_debug)
                        fprintf(stderr, "set hostonset (b) %ld\n",
                                hafd->h_hostonset);
                    if (sd_debug)
                        fprintf(stderr,
                                "after: head %ld, backhead %ld\n",
                                newhead, newbackhead);
                    didsomething = 1;
                }
                hafd->h_skipbytes = 0;
            }
        }
    }
}

#ifdef NEXTSTEP
void sd_pollslot(slotno)
{
    t_hasyncfd *hafd = sd_asyncfds[slotno].h;
    t_asyncfd *cpfd = PROTO->p_asyncfd;
    int i;
    mmap_slotno(slotno);
    for (i = 0; i < MAXAFD; i++, hafd++, cpfd++)
        sd_dopoll(hafd, cpfd);
    if (sd_debug && didsomething)
        fprintf(stderr, "didsomething\n");
}
#endif

sd_pollit()
{
    int i;
    didsomething = 0;
#ifdef NEXTSTEP
    for (i = 0; i < MAXSLOTNO; i++)
    {
        if (sd_slots[i])
            sd_pollslot(i);
    }
#endif
#ifdef __unix
    sd_dopoll(&sd_hasyncfd, sd_asyncfd);
#endif
    return (didsomething);
}

main(argc, argv)
char **argv;
{
#ifdef NEXTSTEP
    char *slotstring;
    int th1, th2, th3;
    argc--;
    argv++;
    if (argc < 5
        || sscanf(argv[0], "%d", &sd_serverno) < 1 || sd_serverno <= 0
        || sscanf(argv[1], "%d", &th1) < 1 || th1 < 0
        || sscanf(argv[2], "%d", &th2) < 1 || th2 < 0
        || sscanf(argv[3], "%d", &th3) < 1 || th3 < 0)
    {
        fprintf(stderr, "usage: max-sounddisk <serverno> <throttle1> \
      \n\t<throttle2> <throttle3> <slot> ...\n");
        exit(1);
    }
    mmap_open(3);
    mmap_setthrottle(th1, th2, th3);
    slotstring = argv[4];
    while (1)
    {
        int thatslot;
        if (sscanf(slotstring, "%d", &thatslot) < 1)
            break;
        else if (thatslot < 0 || thatslot >= MAXSLOTNO)
        {
            fprintf(stderr, "sounddisk: bad slot list\n");
            exit(1);
        }
        else
            sd_slots[thatslot] = 1;
        while (*slotstring && *slotstring != ',')
            slotstring++;
        if (*slotstring)
            slotstring++;
    }
    if (setpriority(PRIO_PROCESS, 0, -17) < 0)
        fprintf(stderr, "max-sounddisk: warning: couldn't set priority\n");
    seteuid(getuid());
#endif
#ifdef __unix

/*
 * On the SGI we have been passed an integer,  "shmid",  which we can attach
 * to to get the shared memory.
 */
    sd_serverno = 1;
    if (argc < 2 || sscanf(argv[1], "%d", &sd_shmid) < 1)
    {
        fprintf(stderr, "max-sounddisk: bad arg\n");
        exit(1);
    }
    if ((sd_shmaddr = shmat(sd_shmid, 0, 0)) == (void *) -1)
    {
        fprintf(stderr, "max-sounddisk: shared memory error\n");
        perror("shmat");
        exit(1);
    }
    sd_asyncfd = ((t_asyncfd *) (sd_shmaddr))->afd_self;
    sd_memoffset = (char *) sd_shmaddr - (char *) sd_asyncfd;
#endif

    sd_init();
    while (!sd_stdinready())
    {
        int foo;
        if (sd_debug)
            fprintf(stderr, "poll %d\n", sd_shmid);
        sd_open = 0;
        foo = sd_pollit();
        if (!foo)
        {
            if (sd_debug)
                sd_minisleep(sd_open ? 300 : 2000);
            else
                sd_minisleep(sd_open ? 50 : 200);
        }
    }

#ifdef __unix
    /* detach the shared memory */
    if (shmdt(sd_shmaddr) < 0)
        perror("shmdt");

    /* remove the ID -- should I really do this here!? */
    if (shmctl(sd_shmid, IPC_RMID, 0) < 0)
        perror("shmctl");
#endif
    if (sd_debug)
        fprintf(stderr, "exit\n");
    exit(0);
}
